#!/usr/bin/env python

import sys
import getopt
import os 
import string 

# ---------------------------------------------------------------------- 
# Default versions. They can be changed here, or (preferred) directly from 
# the command line, e.g.: 
#           source configure --hepmc-version=2.06.03
# ---------------------------------------------------------------------- 
versions = {
  'cascade-version':          '2.0.1',
  'clhep-version':            '1.9.4.7',
  'hepmc-version':            '2.06.05',
#  'heppdt-version':           '3.04.01',
  'fastjet-version':          '3.0.3',
  'gsl-version':              '1.10',
  'pythia8-version':          '170',
  'pythia6-version':          '423.2',
  'alpgen-version':           '2.1.3d.2',
  'herwigpp-version':         '2.6.1',
  'herwig-version':           '6.520.2',
  'jimmy-version':            '4.31.3',
  'thepeg-version':           '1.8.1',
  'lhapdf-version':           '5.8.2',
  'tauola-version':           '28.121.2',
  'photos-version':           '215.5',
  'prefix':                   ''
}

# auxiliary stuff 
colours = { 'default':   '',
            'blue':      '\x1b[01;34m',
            'cyan':      '\x1b[01;36m',
            'green':     '\x1b[01;32m',
            'red':       '\x1b[01;05;37;41m',
            'end':       '\x1b[00m' }

# 
envVarsFile   = open( 'setupEnvVariables.sh', 'w' )
gccSetupFile  = open( 'setupGccVersion.sh', 'w' )

# ---------------------------------------------------------------------- 
def main():
  processOptions()    
 
  envVarsFile.write( '#!/bin/bash\n' ) 
  gccSetupFile.write( '#!/bin/bash\n' ) 

    # location's specific setup  
  location, curr_dir = detectLocation()
  envVarsFile.write( 'export MYCLASS_PATH=' + curr_dir + '\n' ) 
  location = location.rstrip( '\n' ) 

  if location == 'desy.de':
    prefix, sysname = initializeDesySetup()
  elif location == 'naf.desy.de':  
    prefix, sysname = initializeNafSetup()
  else:  
    prefix, sysname = initializeCernSetup()

  print "Prefix:", prefix, ", sysname:", sysname
  initializeStdLibs( prefix, sysname ) 
  
  print 'Done!' 
  
# ---------------------------------------------------------------------- 
# Process command line options  
# ---------------------------------------------------------------------- 
def processOptions():
  options     = ''
  longOptions = [] 
  for op in versions.keys():
    longOptions.append( op + '=' )    # options which require arguments need '='
  
    # add '--help' option
  longOptions.append( 'help' )

    # parse command line options
  try:
    optlist, args = getopt.getopt( sys.argv[1:], options, longOptions )
  except getopt.GetoptError, err:
    print str( err )
    print "For help use --help"
    sys.exit( 2 )

  for op, ar in optlist:
    if op.startswith( '--' ):   # remove '--' from the beginning
      op = op.strip( '-' )
    if op in versions.keys():
      versions[ op ] = ar
      print colours[ 'blue' ]+"Overwriting default option '", op, "' with value:", ar, colours['end']
      if op == 'hepmc-version':
        print colours['blue']+"Be aware that other libraries could be compiled with different version of HepMC!", colours['end']
    if op == 'help':
      showHelp()        
      sys.exit( 0 )

# ---------------------------------------------------------------------- 
# DESY specific setup
# ---------------------------------------------------------------------- 
def initializeDesySetup():
  print colours[ 'blue' ] + "[ Initialize DESY setup ]" + colours[ 'end' ]
  
  #sysNameInfo = os.popen( 'source check_platform_desy' ).readlines()
  #sysname = sysNameInfo[0]
  #print 'Systsssssem:', sysname,
  #envVarsFile.write( 'export SYSNAME=' + sysname + '\n' )

    # determine system's details
  sysArch = 0
  sysDsc  = ''
  sysName = ''
  sysCmd = os.popen( 'fs sysname' ).readline()
    # == SL4
  if contains( sysCmd, 'i586_rhel40' ) or contains( sysCmd, 'i386_linux26' ):
    sysArch = 32
    sysDsc  = 'SL4'
    sysName = 'i586_rhel40'
  elif contains( sysCmd, 'amd64_rhel40' ) or contains( sysCmd, 'amd64_linux26' ):
    sysArch = 64
    sysDsc  = 'SL4'
    sysName = 'amd64_rhel40'
    # == SL5
  if contains( sysCmd, 'i586_rhel50' ):
    sysArch = 32
    sysDsc  = 'SL5'
    sysName = 'i586_rhel50'
  elif contains( sysCmd, 'amd64_rhel50' ):
    sysArch = 64
    sysDsc  = 'SL5'
    sysName = 'amd64_rhel50'

  print 'System:', sysName, '(', sysDsc, '), architecture:', sysArch
  envVarsFile.write( 'export SYSNAME=' + sysName + '\n' )

    # SL4 specific settings
  if sysDsc == 'SL4':
    print 'Initialize SL4 settings'
    envVarsFile.write( 'export CXX="/usr/bin/g++"\n' )
    envVarsFile.write( 'export F77="/usr/bin/g77"\n' )
    envVarsFile.write( 'export FLIBS="-lfrtbegin -lg2c"\n' )
    # SL5 specific settings
  elif sysDsc == 'SL5':
    print 'Initialize SL5 settings'
    envVarsFile.write( 'export CXX="/usr/bin/g++44"\n' )
    envVarsFile.write( 'export F77="/usr/bin/gfortran44"\n' )
    envVarsFile.write( 'export FLIBS="-lgfortran -lgfortranbegin"\n' )
  else:
    print 'Unknown system name:',sysname,
 
    # prefix 
  if len( versions[ 'prefix' ] ) == 0:
    prefix = '/afs/desy.de/group/alliance/mcg/public'
  else:  
    prefix = versions[ 'prefix' ] 
  envVarsFile.write( 'export PREFIX=' + prefix + '\n' )

    # clhep
  envVarsFile.write( 'export CLHEPversion=' + versions[ 'clhep-version' ] + '\n' );
  clhepDir = prefix + '/CLHEP/' + versions[ 'clhep-version' ] + '/' + sysName   
  print 'CLHEP directory:', clhepDir
  envVarsFile.write( 'export CLHEPdir=' + clhepDir + '\n' );
 
    # cernlib 
  if sysDsc == 'SL5':
    envVarsFile.write( 'export CERNLIB=' + prefix + '/' + sysName + '/lib/cernlib/2006a/lib\n' )
  elif sysDsc == 'SL4':
    envVarsFile.write( 'export CERNLIB=/opt/products/cernlib/2005/lib\n' )

   # Root's initialization
  rootVersion = '5.22.00'    
  print 'ROOT version:', rootVersion
  envVarsFile.write( 'export ROOTver=' + rootVersion + '\n' )
  
  if sysName.find( 'amd64' ) != -1:
    rootSys = '/opt/products/root64/' + rootVersion
    rootLib = rootSys + '/lib64'
  else:
    rootSys = '/opt/products/root/' + rootVersion
    rootLib = rootSys + '/lib'
  envVarsFile.write( 'export ROOTSYS=' + rootSys + '\n' )
  envVarsFile.write( 'export ROOTLIB=' + rootLib + '\n' )
  envVarsFile.write( 'export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$ROOTLIB\n' )
  envVarsFile.write( 'export PATH=$ROOTSYS/bin:$PATH\n' )
      
  envVarsFile.write( 'export PDFPATH=' + prefix + '/MCGenerators/cascade/' + versions[ 'cascade-version' ] + '/data\n' )

  return prefix, sysName

# ---------------------------------------------------------------------- 
# NAF specific setup (TODO!)
# ---------------------------------------------------------------------- 
def initializeNafSetup():
  print '== Initialize NAF setup =='

  # determine system
  sysNameInfo = os.popen( 'source check_platform_desy' ).readlines()
  sysname = sysNameInfo[0]
  print 'System:', sysname,
  envVarsFile.write( 'export SYSNAME=' + sysname + '\n' )

  # prefix 
  if len( versions[ 'prefix' ] ) == 0:
    prefix = '/afs/desy.de/group/alliance/mcg/public'
  else:  
    prefix = versions[ 'prefix' ] 
  envVarsFile.write( 'export PREFIX=' + prefix + '\n' )

  # Root's initialization
  rootVersion = '5.22.00'    
  print 'ROOT version:', rootVersion
  envVarsFile.write( 'export ROOTver=' + rootVersion + '\n' )
  
  if sysname.find( 'amd64' ) != -1:
    rootSys = '/afs/naf.desy.de/products/root/' + rootVersion
    rootLib = rootSys + '/lib'
  else:
    print 'No 32 bit ROOT version at NAF!'
    sys.exit( 2 )
  envVarsFile.write( 'export ROOTSYS=' + rootSys + '\n' )
  envVarsFile.write( 'export ROOTLIB=' + rootLib + '\n' )
  envVarsFile.write( 'export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$ROOTLIB\n' )
  envVarsFile.write( 'export PATH=$ROOTSYS/bin:$PATH\n' )
      
  # clhep...
  envVarsFile.write( 'export CLHEPversion=' + versions[ 'clhep-version' ] + '\n' );
  clhepDir = prefix + '/CLHEP/' + versions[ 'clhep-version' ] + '/' + sysname   
  print 'CLHEP dir:', clhepDir,
  envVarsFile.write( 'export CLHEPdir=' + clhepDir + '\n' );
  
  envVarsFile.write( 'export CERNLIB=/opt/products/cernlib/2005/lib\n' )
  envVarsFile.write( 'export PDFPATH=' + prefix + '/MCGenerators/cascade/' + versions[ 'cascade-version' ] + '/data\n' )

  return prefix, sysname

# ---------------------------------------------------------------------- 
# CERN specific setup
# ---------------------------------------------------------------------- 
def initializeCernSetup():
  print colours[ 'blue' ] + "[ Initialize CERN setup ]" + colours[ 'end' ]
  
  # determine system
  curr_dir1 = os.popen( 'pwd' ).read()
  print 'Current working directory1:', curr_dir1
  sysNameInfo = os.popen( 'source ./check_platform' ).readlines()
  print 'sysnameinfo ',sysNameInfo
  sysname = sysNameInfo[1].split()[2]
  #sysname = sysNameInfo[0]
  #sysname = sysname.replace( "\n", "" )
  print 'System:', sysname
  envVarsFile.write( 'export SYSNAME=' + sysname + '\n' )

  tmpTab = sysname.split( '-' )   # e.g. x86_64-slc5-gcc43-opt
  if len( tmpTab ) > 1:
    arch = tmpTab[ 0 ]                # e.g. x86_64
    slcVer = tmpTab[ 1 ]
    print 'Architecture:', arch
    print 'SLC version:', slcVer 
    if slcVer == 'slc5':
      ###gccSetup = '/afs/cern.ch/sw/lcg/external/gcc/4.3.2/' + arch + '/setup.sh\n'
      gccSetup = '/afs/cern.ch/sw/lcg/external/gcc/4.3.5/' + arch + '/setup.sh\n'
      print 'Gcc setup file:', gccSetup
      gccSetupFile.write( 'source %s' % gccSetup )
#      envVarsFile.write( 'source /afs/cern.ch/sw/lcg/external/gcc/4.1.2/' + arch + '/setup.sh\n' )

    elif slcVer == 'slc6':
      gccSetup = '/afs/cern.ch/sw/lcg/external/gcc/4.6/' + arch + '-slc6''/setup.sh\n'
      #/afs/cern.ch/sw/lcg/external/gcc/4.6.2/x86_64-slc6/setup.sh
      print 'Gcc setup file:', gccSetup
      gccSetupFile.write( 'source %s' % gccSetup )


  # determine GCC version
  gccVersion = os.popen( 'g++ --version' ).readline().split()[2]
  if gccVersion.startswith( '4' ):
    envVarsFile.write( 'export GFORTRAN=1\n' )
    print 'GFORTRAN = 1'
    envVarsFile.write( 'export F77="/usr/bin/gfortran44"\n' )
    envVarsFile.write( 'export FLIBS="-lgfortran -lgfortranbegin"\n' )
  else:
    print 'GFORTRAN = 0'
    envVarsFile.write( 'export F77="/usr/bin/g77"\n' )
    envVarsFile.write( 'export FLIBS="-lfrtbegin -lg2c"\n' )
  print 'GCC version:', gccVersion
 
  # prefix 
  if len( versions[ 'prefix' ] ) == 0:
    prefix = '/afs/cern.ch/sw/lcg/external'
  else:  
    prefix = versions[ 'prefix' ] 
  envVarsFile.write( 'export PREFIX=' + prefix + '\n' )

  # Root's initialization
  rootVersion = '5.34.00'    
  print 'ROOT version:', rootVersion
  envVarsFile.write( 'export ROOTver=' + rootVersion + '\n' )
  rootSys = '/afs/cern.ch/sw/lcg/app/releases/ROOT/' + rootVersion + '/' + sysname + '/root'
  print 'ROOT path:', rootSys
  envVarsFile.write( 'export ROOTSYS=' + rootSys + '\n' )
  rootLib = rootSys + '/lib'
  print 'ROOT lib:', rootLib
  envVarsFile.write( 'export ROOTLIB=' + rootLib + '\n' )
  envVarsFile.write( 'export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:' + rootLib + '\n' )
  envVarsFile.write( 'export PATH=' + rootSys + '/bin:$PATH\n' )

  # clhep...
#  versions[ 'clhep-version' ] = '2.0.4.4'
  envVarsFile.write( 'export CLHEPversion=' + versions[ 'clhep-version' ] + '\n' );
  clhepDir = prefix + '/clhep/' + versions[ 'clhep-version' ] + '/' + sysname   
  print 'CLHEP dir:', clhepDir
  envVarsFile.write( 'export CLHEPdir=' + clhepDir + '\n' );
  
  envVarsFile.write( 'export CERNLIB=' + prefix + '/cernlib/2006b/' + sysname + '/lib' + '\n' )
  envVarsFile.write( 'export PDFPATH=' + prefix + '/MCGenerators/cascade/' + versions[ 'cascade-version' ] + '/share/data\n' )
  return prefix, sysname

# ---------------------------------------------------------------------- 
# Setup environment variables
# ---------------------------------------------------------------------- 
def initializeStdLibs( prefix, sysname ):
  hepmc_suff = '' 
  #pythia8_hepmc_suff = '' 
  #if ( versions[ 'hepmc-version' ] == '2.06.03' ):      # use generators compiled with the new version of HepMC,
  pythia8_hepmc_suff = '_hepmc' + versions['hepmc-version']                  # currently (Nov/2010) available only for Pythia8 at Cern...
  
  herwigpp_hepmc_suff = '_hepmc' + versions['hepmc-version']
  thepeg_hepmc_suff = '_hepmc' + versions['hepmc-version']
  
  envVarsFile.write( 'export HepMCdir=' + prefix + '/HepMC/' + versions[ 'hepmc-version' ] + '/' + sysname + '\n' )
#  envVarsFile.write( 'export HepPDTdir=' + prefix + '/HepPDT/' + versions[ 'heppdt-version'] + '/' + sysname + '\n' )

  envVarsFile.write( 'export FastJetdir=' + prefix + '/fastjet/' + versions[ 'fastjet-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export GSLdir=' + prefix + '/GSL/' + versions[ 'gsl-version'] + '/' + sysname + '\n' )
  
  envVarsFile.write( 'export Pythia8dir=' + prefix + '/MCGenerators' + pythia8_hepmc_suff + '/pythia8/' + versions[ 'pythia8-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export Pythia8DATA=' + prefix + '/MCGenerators' + pythia8_hepmc_suff + '/pythia8/' + versions[ 'pythia8-version'] + '/share/xmldoc\n' )
  
  envVarsFile.write( 'export Pythia6dir=' + prefix + '/MCGenerators' + hepmc_suff + '/pythia6/' + versions[ 'pythia6-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export Alpgendir=' + prefix + '/MCGenerators' + hepmc_suff + '/alpgen/' + versions[ 'alpgen-version'] + '/' + sysname + '\n' )
  
  envVarsFile.write( 'export Herwigppdir=' + prefix + '/MCGenerators' + herwigpp_hepmc_suff + '/herwig++/' + versions[ 'herwigpp-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export ThePEGdir=' + prefix + '/MCGenerators' + thepeg_hepmc_suff + '/thepeg/' + versions[ 'thepeg-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export Herwigdir=' + prefix + '/MCGenerators' + hepmc_suff + '/herwig/' + versions[ 'herwig-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export Jimmydir=' + prefix + '/MCGenerators' + hepmc_suff + '/jimmy/' + versions[ 'jimmy-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export LHAPDFdir=' + prefix + '/MCGenerators' + hepmc_suff + '/lhapdf/' + versions[ 'lhapdf-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export LHAPDFsets=' + prefix + '/MCGenerators' + hepmc_suff + '/lhapdf/' + versions[ 'lhapdf-version'] + '/share/PDFsets\n' )
  envVarsFile.write( 'export LHAPATH=' + prefix + '/MCGenerators' + hepmc_suff + '/lhapdf/' + versions[ 'lhapdf-version'] + '/share/PDFsets\n' )
  envVarsFile.write( 'export Cascadedir=' + prefix + '/MCGenerators' + hepmc_suff + '/cascade/' + versions[ 'cascade-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export Tauoladir=' + prefix + '/MCGenerators' + hepmc_suff + '/tauola/' + versions[ 'tauola-version'] + '/' + sysname + '\n' )
  envVarsFile.write( 'export Photosdir=' + prefix + '/MCGenerators' + hepmc_suff + '/photos/' + versions[ 'photos-version'] + '/' + sysname + '\n' )

    # additional environment variables
  envVarsFile.write( 'export Herwigppversion=' + versions[ 'herwigpp-version'] + '\n' )
  
    # add to LD_LIBRARY_PATH
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Pythia8dir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Pythia8DATA\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Pythia6dir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Alpgendir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$GSLdir/lib/\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$HepMCdir/lib\n" )
#  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$HepPDTdir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$CLHEPdir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$FastJetdir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$LHAPDFdir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Herwigppdir/lib/Herwig++\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Herwigdir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Jimmydir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$ThePEGdir/lib/ThePEG\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Cascadedir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Tauoladir/lib\n" )
  envVarsFile.write( "LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$Photosdir/lib\n" )
  envVarsFile.write( "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$MYCLASS_PATH/lib\n" )

  return  

# ---------------------------------------------------------------------- 
def detectLocation():
  loc = os.popen( 'hostname -d' ).read()
  print 'Location:', loc,
  curr_dir = os.popen( 'pwd' ).read()
  print 'Current working directory:', curr_dir,
  return loc, curr_dir

# ---------------------------------------------------------------------- 
def contains( theString, theQueryValue ):
  return theString.find( theQueryValue ) > -1 

# ---------------------------------------------------------------------- 
def showHelp():
  print 'configure.py - HepMCAnalysis tool configuration script (Version: 0.1)\n'
  print 'Usage: source configure [arguments]\n'
  print 'Arguments:'
  for arg in versions.keys():
    print '  --' + arg 
  print '\nTo specify different version of a generator use command like:'
  print '  source configure --pythia8-version=135\n'

# ---------------------------------------------------------------------- 
# Execution starts here  
# ---------------------------------------------------------------------- 
if __name__ == "__main__":
  main()
  envVarsFile.close() 
  gccSetupFile.close()

