//---------------------------------------------------------------------- 
// Create the web display with:
//   root plot_new.C+
//
//   This will start ROOT and execute the macro plot.C to create the
//   web-display.
//
//   In the target directory a sub-dir for each analysis will be created.
//---------------------------------------------------------------------- 

// STL 
#include <vector>
#include <string>
#include <fstream>
#include <iostream>
#include <sstream>

// Root 
#include <TROOT.h>
#include <TPaveText.h>
#include <TSystem.h>
#include <TCanvas.h>
#include <TFile.h>
#include <TText.h>
#include <TStyle.h>
#include <TKey.h>
#include <TH1.h>
#include <TLegend.h>
#include <TDirectory.h>
//#include <TDirectoryFile.h>

// local headers
#include "atlas_style.C"

//---------------------------------------------------------------------- 
// Forward declarations 
//---------------------------------------------------------------------- 
bool Initialize();
bool ReadFileList( const std::string &fileName );
bool ScanFiles();
bool PlotHistograms();
void GenerateHtmlPage( const std::string &targetDir );

void ExitRoot();
void CleanUp();
bool IsComment( const std::string &line );

//---------------------------------------------------------------------- 
// Global variables (Note: names start with 'g') 
//---------------------------------------------------------------------- 
std::vector< TFile* > gRootFiles;
std::vector< std::string > gHistoExt;
std::vector< std::string > gLogscaleIdentifier;
std::vector< std::string > gHistoNames;
std::vector< std::string > gGeneratorLabel;
std::vector< std::string > gRootSubdirs;
std::vector< double > gCrossSections;
std::vector< double > gHistoMaxValues;
std::vector< int > gNumberOfEvents;

std::string gBaseDirectory;
std::string gPwd;

//---------------------------------------------------------------------- 
// Setup part, please modify these Parameters/Variables according to your needs
//---------------------------------------------------------------------- 
// if this flag is set to true, than the target directory will be created, if it does not exist 
bool gCreateTargetDir = true;
// if this flag is set to true, HTML page will be created
bool gGenerateHTMLpage = true;

std::string gPlotListFileName( "plotFiles.list" );

// ---------------------------------------------------------------------- 
// ==> Execution starts here
// ---------------------------------------------------------------------- 
void plot_new()
{
  using namespace std;
  
  gROOT->Reset();
  Initialize();
  
  if ( !ReadFileList( gPlotListFileName ) ) {
    cerr << "Error! cannot read file list: " << gPlotListFileName << endl;
    ExitRoot();
    return;
  }

  gSystem->cd( gPwd.c_str() );  // change back to the working directory

  if ( !ScanFiles() ) {
    cerr << "Error! error during file scan" << endl;
    ExitRoot();
    return;
  }

  PlotHistograms();
//  GenerateHtmlPage ( "~/scratch0/ww_tmp_Tau" );

} // eof plot_new() 

// ---------------------------------------------------------------------- 
bool Initialize()
{
  using namespace std;

  // enforce ATLAS style
  gROOT->ProcessLine( ".L atlas_style.C" );
  useAtlasStyle();

  // Set Output file format ... you can specify several formats (e.g. gif (wabpage) and eps (paper))
  gHistoExt.push_back( "eps" );
  gHistoExt.push_back( "gif" );
  gHistoExt.push_back( "png" );
  gHistoExt.push_back( "tiff" );

  // Set the base directory ... The histogramms will be saved in this directory
  gPwd = gSystem->pwd();

  // If the name contains this string than 
  // logscale will be enabled
  gLogscaleIdentifier.push_back( "_logscale" );

  return true;
}

// ---------------------------------------------------------------------- 
// Read list of ROOT files to be processed
// ---------------------------------------------------------------------- 
bool ReadFileList( const std::string &fileName )
{
  using namespace std;
  
  ifstream myfile( fileName.c_str() );
  
  cout << "Reading list of ROOT files from: " << fileName << endl;
  if ( !myfile.is_open() ) {
    cerr << "Error: unable to open file: " << fileName << endl;
    return false;
  }

  string line;

  // read line by line of the file 
  while ( !myfile.eof() ) {
    // write read lines into string line
    getline( myfile, line );

    // check for Comments. If the line starts with # than skip the line 
    if ( IsComment( line ) ) {
      continue;
    }
	    
    cout << line << endl;

	  // now split the line into the subwords
	  // and write the substrings into vector
	  istringstream instring( line );
	  string string1, string2, string3;

    // split the line again but using as numbers to read in
	  // the crosssections 
	  istringstream instring2( line );
    string temp;
	  double mydouble1 = 0., mydouble2 = 0.;

    instring >> string1 >> string2 >> string3;
    instring2 >> temp >> temp >> temp >> mydouble1 >> mydouble2;

    if( ( string1 == "->" ) || ( string1 == "in" ) ) {
      // check wether the file is a root-file
      string::size_type posExt = string3.find( ".root" );         // should end with .root
      if ( posExt == string::npos ) {
        continue;     // not a Root file
      }

      // set global variables
      gRootFiles.push_back( new TFile( string3.c_str() ) );       // ?? 
      gGeneratorLabel.push_back( string2 );
      gCrossSections.push_back( mydouble1 );
      gNumberOfEvents.push_back( static_cast< int >( mydouble2 ) );

      if( !gRootFiles.back()->IsOpen() ) {
        cerr << "Error: file " << string3 << " could not be opened... skip!" << endl;
        return false;
      }
    } else if( ( string1 == "<-" ) || ( string1 == "out" ) ) {
      gBaseDirectory = string2;
      continue;
	  }
  }   // end of file reading

  myfile.close();

  return true;
}   // end of ReadFileList()

// ---------------------------------------------------------------------- 
// 
// ---------------------------------------------------------------------- 
bool ScanFiles()
{
  using namespace std;
  
  cout << "First scan through the files to get the needed Information ..." << endl;
  
  size_t nFiles = gRootFiles.size();
  cout << "# of files: " << nFiles << endl;

  // need to scan rootfiles for subdirs and histograms
  TKey  *key_dir = 0,
        *key_histo = 0;

  // Loop  over the files
  for ( size_t files = 0; files < nFiles; files++ ) {
    TFile *&currentFile = gRootFiles.at( files );

    // list the files
    cout << "File " << files + 1 << ": " <<  currentFile->GetName() << endl;
    currentFile->cd();
    
    // get the iterator for the subdirectories in the file
    TIter nextkey( currentFile->GetListOfKeys() );

    // loop over the TDirectories
    while ( key_dir = ( TKey* )nextkey() ) {
      TDirectory *currentrootdir = ( TDirectory* )key_dir->ReadObj();
//      TDirectoryFile *currentrootdir = dynamic_cast< TDirectoryFile* >( key_dir->ReadObj() );
      string classname = currentrootdir->ClassName();
      if( classname != string( "TDirectoryFile" ) ) {
        continue;
      }

      string currentrootdirname = currentrootdir->GetName();
	
      // get the name of the TDirectory
      bool direxists = false;
      for( size_t dir = 0; dir < gRootSubdirs.size(); ++dir ) {
	      if ( gRootSubdirs.at( dir ) == currentrootdirname ) {
          direxists = true;
          break;
        }
      }

      if ( !direxists ) {
        gRootSubdirs.push_back( currentrootdirname );
      }
	
      // Loop over the histograms in the TDirectory
      TIter nextkey2( currentrootdir->GetListOfKeys() );
      while ( key_histo = (TKey*)nextkey2() ) {
	      TH1D* currenthisto = dynamic_cast< TH1D* >( key_histo->ReadObj() );
  
        if ( currenthisto == 0 ) {
          continue;      /// not a histogram -> skip
        }
  
//        string classname2 = currenthisto->ClassName();
//        if( classname2 != string( "TH1D" ) ) {
//          continue; 
//        }
	    
        // Number of bins of the current histogram
        int numberOfBins = currenthisto->GetNbinsX();
        for( int nbin = 0; nbin <= numberOfBins; ++nbin ) {
          double BinContent = currenthisto->GetBinContent( nbin );
          double BinError = currenthisto->GetBinError( nbin );
          
          double ratio = gCrossSections.at( files ) / gNumberOfEvents.at( files );
          double NewBinContent = BinContent * ratio;
          double NewBinError = BinError * ratio; 

          currenthisto->SetBinContent( nbin, NewBinContent );
          currenthisto->SetBinError( nbin, NewBinError );
	        //currenthisto->Sumw2();
	      }

        // Name of the current histogram
        string currenthistoname = currenthisto->GetName();

        // Maximum of the current histogram
        double currentmax = currenthisto->GetMaximum();

        // check wether histogramname already filled in vector
        bool histoexists = false;
        for( size_t histo = 0; histo < gHistoNames.size(); ++histo ) {
          if( gHistoNames.at( histo ) != currenthistoname ) {
            continue; 
          }
          histoexists = true;
		
          // for histograms with same name (to be compared) set the maximum to the maximum of both histograms
          if ( currentmax > gHistoMaxValues[ histo ] ) {
            gHistoMaxValues[ histo ] = currentmax;
          }
		    
		      // if histogramname found than, quit the loop 		        
      		if( histoexists ) break;
  	    }
	    
        // if it is not found than append name and maximum to the vector
        if( !histoexists ) {
          gHistoNames.push_back( currenthistoname );
          gHistoMaxValues.push_back( currentmax );
        }
	    
//        delete currenthisto;    // ??? 
      }     // loop over histograms
//	    delete currentrootdir;      // ????
    }   // end of loop over directories
  }   // end of loop over files

  cout << "# of histo names: " << gHistoNames.size() << endl;

  return true;
}   // end of ScanFiles()

// ---------------------------------------------------------------------- 
// 
// ---------------------------------------------------------------------- 

// Kolmogorov test minimal value
Double_t kt_min_result=0.1;
Double_t kt_medium_result=0.7;

// Default fill for First distribution drawn
Double_t fillColor=0;

// scale y-range with maximum
Double_t scale_y=1.3;

// constructor options for TCanvas
Int_t wtopx=200;
Int_t wtopy=10;
Int_t ww=600;
Int_t wh=400;

// TLegend Properties
Double_t tl_x1=0.775;
Double_t tl_x2=0.975;
Double_t tl_y1=0.9;
Double_t tl_y2=0.92;
Double_t tl_ystep=0.05;
Double_t tl_textsize=0.035;

//TPaveText Coordinates Pavetext1
Double_t pt1_x1=0.505;
Double_t pt1_x2=0.765;
Double_t pt1_y1=0.8;
Double_t pt1_y2=0.92;
Double_t pt1_textsize=0.03;

//TPaveText Coordinates Pavetext2
Double_t pt2_x1=0.235;
Double_t pt2_x2=0.495;
Double_t pt2_y1=0.75;
Double_t pt2_y2=0.92;
Double_t pt2_ystep=0.0;
Double_t pt2_textsize=0.03-0.01*( gRootFiles.size() - 2 );

TCanvas* gCanvas = 0; 

// ---------------------------------------------------------------------- 
//
// ---------------------------------------------------------------------- 
bool PlotHistograms()
{
  using namespace std;
  cout << "Now actually plot the histograms... " << endl;

  // create a TCanvas to generate the histograms
  //gCanvas = new TCanvas( "gCanvas", "Jet Validation", wtopx, wtopy, ww, wh );
  TCanvas *c1 = new TCanvas( "c1", "Jet Validation", wtopx, wtopy, ww, wh );
//  gCanvas->cd( 1 );

  string  command, 
          histo_filename,
          targetDir;

  // Loop over the directories
  for( size_t dir = 0; dir < gRootSubdirs.size(); ++dir ) {
    cout << gRootSubdirs.at( dir ) << endl;
    targetDir = gBaseDirectory + "_" + gRootSubdirs.at( dir );

    // check wether the directory exists
    if( !gSystem->cd( targetDir.c_str() ) ) {
      cout << "Directory " << targetDir << " does not exist! Will be created ..." << endl;
      gSystem->mkdir( targetDir.c_str() );
      cout << "... done!" << endl;

    } else {
      // if directory exists, than clean up existing files.
      // They will be overwritten anyway, but do not want to mix up old and new files
      command = "rm -rf ";

      for ( size_t ext = 0; ext < gHistoExt.size(); ext++ ) {
	      command += targetDir + "/*." + gHistoExt.at( ext ) + " ";
      }

      gSystem->Exec( command.c_str() );
      cout << "Execute the command: " << command << endl;
    }

    // Loop over the histograms (create image files with the histograms)
    for( size_t histo = 0; histo < gHistoNames.size(); histo++ ) {
      int colour = 1; 
      int lstyle = 1;

      // Define the Legend and set the Properties
      TLegend* mylegend = new TLegend( tl_x1, tl_y1 - gRootFiles.size() * tl_ystep, tl_x2, tl_y2 );
      mylegend->SetHeader( "Generator Validation" );
      mylegend->SetTextSize( tl_textsize );
      mylegend->SetFillColor( kWhite );

      bool histoexists = false;
      bool logscale = false;

      TH1D *firsthisto = 0;

      // create TPaveText to show first Generator statistics
      TPaveText* pt1_text = new TPaveText( pt1_x1, pt1_y1, pt1_x2, pt1_y2, "NDC" );
      pt1_text->SetTextSize( pt1_textsize );
      pt1_text->SetFillColor( kWhite );

      // create TPaveText to show other Generators statistics
      TPaveText* pt2_text = new TPaveText(pt2_x1,pt2_y1,pt2_x2,pt2_y2,"NDC");
      pt2_text->SetTextSize( pt2_textsize );
	    
      //currently set on white until Kolmogorov test is fully understood
      pt2_text->SetFillColor( kWhite );
      
	    // Loop over the inputfiles
      for( size_t files = 0; files < gRootFiles.size(); ++files ) {
        if( ( colour == 1 ) || ( colour == 3 ) || ( colour == 5 ) ) colour++;      // ????? 
        if( ( colour == 1 ) || ( colour == 3 ) || ( colour == 5 ) ) colour++;
        if( ( colour == 1 ) || ( colour == 3 ) || ( colour == 5 ) ) colour++;

        gRootFiles[ files ]->cd();
        TDirectory *currentrootdir = (TDirectory *)gRootFiles[ files ]->Get( gRootSubdirs[ dir ].c_str() );
        
        TH1D *currenthisto = dynamic_cast< TH1D* >( currentrootdir->Get( gHistoNames[ histo ].c_str() ) );
        if( !currenthisto ) {
          continue;
        }

  	    // according to the name the Canvas will be set to logscale
        if( currenthisto->GetEntries() ) {
          for( size_t lscale = 0; lscale < gLogscaleIdentifier.size(); lscale++ ) {
  		      if ( gHistoNames[ histo ].find( gLogscaleIdentifier[ lscale ], 0 ) != string::npos ) { 
              logscale = true;
            }
          }
        }
	    
  	    histoexists = true;
	    
	      TPaveText *temp_pavetext = 0;
	      TText *textline = 0;
	      char temptext[100];
  	    //stringstream temptext;

  	    // Draw the histograms in one Canvas

	      // Number of bins of the current histogram
	      int numberOfBins = currenthisto->GetNbinsX();
	      for(int nbin = 0; nbin <= numberOfBins; nbin++ ){
  	      double BinContent = currenthisto->GetBinContent( nbin );
	        double BinError = currenthisto->GetBinError( nbin );

	        double NewBinContent = BinContent * gCrossSections.at( files ) / gNumberOfEvents.at(files );
	        double NewBinError = BinError * gCrossSections.at( files ) / gNumberOfEvents.at( files );

  	      currenthisto->SetBinContent( nbin, NewBinContent );
	        currenthisto->SetBinError( nbin, NewBinError );
	      }

  	    currenthisto->SetLineColor( colour );
	      currenthisto->SetLineStyle( lstyle++ );

      c1->cd();

	      if ( files == 0 ) { 
		      firsthisto = currenthisto;
		      temp_pavetext = pt1_text;

		      currenthisto->GetYaxis()->SetRangeUser( 0.1, TMath::Max( gHistoMaxValues.at( histo ) * 1.0, 0.1 * logscale) * 
              scale_y * ( 1 + logscale * 50 ) );

          currenthisto->SetFillColor( static_cast< int >( fillColor ) );
          currenthisto->Draw( "HIST" );
        } else {
          currenthisto->Draw( "HIST SAME" );
          temp_pavetext = pt2_text;
  	    }

  	    // fill Legend
	      mylegend->AddEntry( currenthisto, gGeneratorLabel[ files ].c_str(), "l" );

	      // Fill PaveText
        sprintf(temptext,"%s: ", gGeneratorLabel[ files ].c_str() );
        textline=temp_pavetext->AddText(temptext);
	      textline->SetTextAlign(11);

	      sprintf(temptext,"   Entries=%d", int( currenthisto->GetEntries() ) );
  	    textline=temp_pavetext->AddText(temptext);
	      textline->SetTextAlign(11);

  	    sprintf(temptext,"   #bar{x}=%.2g, RMS=%.2g",currenthisto->GetMean(),currenthisto->GetRMS());
	      textline=temp_pavetext->AddText(temptext);
	      textline->SetTextAlign(11);

  	    sprintf(temptext,"   UF=%.2g, OF=%.2g",currenthisto->GetBinContent(0),currenthisto->GetBinContent(currenthisto->GetNbinsX()+1));
	      textline=temp_pavetext->AddText(temptext);
	      textline->SetTextAlign(11);
  	    textline=temp_pavetext->AddText("");
	      textline->SetTextSize(0.004);

	      colour++;

	      if ( firsthisto == currenthisto) {
          continue;
        }

  	    double  tempresult = 0, 
                tempresult2 = 0;

	      // only make test, if histogram not empty
	      if( !currenthisto->GetEntries() ) {
          continue;
        }

  	    tempresult  = currenthisto->KolmogorovTest( firsthisto );
	      tempresult2 = currenthisto->Chi2Test( firsthisto, "WW CHI2/NDF" );

  	    // If Kolmogorov test returns value less than kv_min, fill TPaveText red
	      //currently set on white until Kolmogorov test is fully understood
	      //if(tempresult<=kt_min_result) temp_pavetext->SetFillColor(kRed);
	      //if(tempresult>kt_min_result && tempresult<=kt_medium_result) temp_pavetext->SetFillColor(kYellow);
  	    if(tempresult<=kt_min_result) temp_pavetext->SetFillColor(kWhite);
	      if(tempresult>kt_min_result && tempresult<=kt_medium_result) temp_pavetext->SetFillColor(kWhite);

  	    textline=temp_pavetext->AddText("Tests:");
	      textline->SetTextAlign(11);
	      sprintf(temptext,"KM: %.3g , #chi^{2}/NDF: %.3g",tempresult,tempresult2);
//	      textline=temp_pavetext->AddText(temptext);
  	    textline->SetTextAlign(11);
	      textline=temp_pavetext->AddText("");
	      textline->SetTextSize(0.005);
      }   // end of loop over input files 
      
    	// draw Legend
	    mylegend->Draw();

    	// draw TPaveText
	    if( gRootFiles.size() > 1 ) {
       pt2_text->Draw();
      }
	    pt1_text->Draw();

	    // save the Canvas in a file ... according to the formats specified at the top of the script
    	if( histoexists ) {
  	    for( size_t ext = 0; ext < gHistoExt.size(); ext++ ) {
	        histo_filename = gBaseDirectory + "_" + gRootSubdirs[ dir ] + "/" + gHistoNames[ histo ] + "." + gHistoExt[ ext ];
	        // for some histograms also plot logscale

	        c1->SetLogy( logscale );
  	      c1->Update();
 	      c1->SaveAs( histo_filename.c_str(), "RECREATE" );
	      }
	    }
	    
//      delete mylegend;
    }   // end of loop over the histograms in directory
    
    // If a HTML page is requested, than copy html-script to the targetdir and run script
    if ( gGenerateHTMLpage ) {
      GenerateHtmlPage( targetDir );
    }
  }   // end of loop over the directories

  return true;
}   // end of PlotHistograms()

// ---------------------------------------------------------------------- 
// 
// ---------------------------------------------------------------------- 
void GenerateHtmlPage( const std::string &targetDir )
{
    // Python script generating HTML page
  const string  makeHtmlScriptName = "mk_html.py";
  const string  utilsName = "pages.py";
    // HTML page template
  const string  overviewPage  = "overview.html",
                indexPage     = "index.html",
                emptyPage     = "empty.html",
                analysisPage  = "analysis.html",
                cssStyle      = "page_style.css";
   
  cout << "Generating HTML page..." << endl;
    
  string command;
    // copy and execute script
  command = "cp " + gPwd + "/" + makeHtmlScriptName + " " + targetDir + "/.";
  cout << "  > execute the command: " << command << endl;
  gSystem->Exec( command.c_str() );
 
  command = "chmod +x " + targetDir + "/" + makeHtmlScriptName;
  gSystem->Exec( command.c_str() );

  command = "cp " + gPwd + "/" + utilsName + " " + targetDir + "/.";
  cout << "  > execute the command: " << command << endl;
  gSystem->Exec( command.c_str() );

  command = "cp " + gPwd + "/" + cssStyle + " " + targetDir + "/.";
  cout << "  > execute the command: " << command << endl;
  gSystem->Exec( command.c_str() );

  command = "cd " + targetDir;
  cout << "  > execute the command : " << command << endl;
  gSystem->Exec( command.c_str() );
  
  command = targetDir + "/" + makeHtmlScriptName;
  cout << "  > start HTML generation... " << command << endl;
  cout << "  > execute the command : " << command << endl;
  gSystem->Exec( command.c_str() );

    // go back to cwd
  command = "cd " + gPwd;
  gSystem->Exec( command.c_str() );
  cout << "Execute the command : " << command << endl;
}

// ---------------------------------------------------------------------- 
// 
// ---------------------------------------------------------------------- 
void CleanUp()
{
  gRootFiles.clear();
  gHistoExt.clear();
  gLogscaleIdentifier.clear();
  gHistoNames.clear();
  gGeneratorLabel.clear();
  gRootSubdirs.clear();
  gCrossSections.clear();
  gHistoMaxValues.clear();
  gNumberOfEvents.clear();
  
  // clean up pointers
  delete gCanvas;
}

// ---------------------------------------------------------------------- 
// A few auxiliary functions
// ---------------------------------------------------------------------- 
void ExitRoot()
{
  std::cout << "Exiting ROOT..." << std::endl;
  gROOT->ProcessLine( ".q" );
}

// ---------------------------------------------------------------------- 
// Return 'true' if the given line is a comment 
// ---------------------------------------------------------------------- 
bool IsComment( const std::string &line )
{
  std::string::size_type pos = line.find( "#" );
  return pos != std::string::npos;
}



