/**
 * Pythia6 example program.
 * 
 * Typical usage:
 *   ./pythia6.exe Process.config
 *
 * Last update (08.2012), Sebastian Piec
 */

#include <iostream>
#include <cassert>

#include "HepMC/PythiaWrapper.h"
#include "HepMC/IO_HEPEVT.h"
#include "HepMC/IO_BaseClass.h"
#include "HepMC/GenEvent.h"
#include "HepMC/SimpleVector.h"
#include "MyPythia6Wrapper.h"

#include <TFile.h>

// HepMCAnalysis headers 
#include "../../include/Configuration.h"
#include "../../include/JetFinder.h"
#include "readpythiaConfigFile.h"

using namespace std;

// ---------------------------------------------------------------------- 
int main(int argc, const char* argv[]) 
{
  if (argc < 2) {
    cerr << "Usage: " << argv[0] << " <Configuration File>" << endl;
    return 1;
  }
 
    // load job's configuration 
  Configuration config; 
  if (!config.read(argv[1])) {
    cerr << " ... skip !" << endl;
    return 1;
  }
 
    // initialize the jet finder 
  JetFinder jetFinder;

  jetFinder.Init(config.max_eta, config.min_pt);
  jetFinder.InitJetFinder(config.coneRadius, config.overlapThreshold, 
      config.jet_ptmin, config.lepton_ptmin, config.DeltaR_lepton_track, 
      config.jetalgorithm);

    // instantiate analyses requested by a user 
  vector<baseAnalysis*> analyses = config.instantiateAnalyses();

  for (vector<baseAnalysis*>::iterator a = analyses.begin(); a != analyses.end(); ++a) {
    (*a)->Init(config.max_eta, config.min_pt);
    (*a)->InitJetFinder(config.coneRadius, config.overlapThreshold, 
        config.jet_ptmin, config.lepton_ptmin, config.DeltaR_lepton_track, config.jetalgorithm);
  }

  pyjets.n = 0;         // ensure dummyness of the next call
  call_pyhepc(1);       // mstu(8) is set to NMXHEP in this dummy call (version >=6.404)
  
  HepMC::HEPEVT_Wrapper::set_max_number_entries(10000);
  HepMC::HEPEVT_Wrapper::set_sizeof_real(8);
  
    // set random number seed (mandatory!) ... this may be overwritten by function readConfigFile
  pydatr.mrpy[1-1] = config.rseed ;
  
  cout << "Configuration parameter list:" << endl;
  
  const vector<string> &fileNames = config.ConfigFileNames;
  for (size_t iFile = 0; iFile < fileNames.size(); ++iFile) {
    readConfigFile(fileNames[iFile]);
  }

    // initialize Pythia 
  call_pyinit("CMS", "p", "p", 7000.0);

  call_pystat(5);
  
    // initialize HepMC 
  HepMC::IO_HEPEVT hepevtio;
 
  ///////////////////////////////////////////////////////////// 
  // Main event loop
  ///////////////////////////////////////////////////////////// 
  for (size_t iEvent = 1; iEvent <= config.nevents; ++iEvent) {
    if (iEvent % 10 == 0) {
      cout << "Processing event " << iEvent << endl;
    }
    
    call_pyevnt();          // generate single event with Pythia6 
    call_pyhepc(1);         // convert common PYJETS to common HEPEVT

    HepMC::GenEvent* event = hepevtio.read_next_event();    // read event from HEPEVT common 
    
    event->set_event_number(iEvent);        // supply additional information to the event
    event->set_signal_process_id(20);

    // ascii_io << evt;     // write the event out to the ascii file
    
      // call jet finder
    vector<fastjet::PseudoJet> inclusiveJets = jetFinder.GetJet(event);

      // run analyses
    bool allFine = true;
    for (vector<baseAnalysis*>::iterator a = analyses.begin(); a != analyses.end(); ++a) {
      assert(*a);
      (*a)->SetJet(&inclusiveJets);

      if ( !(*a)->Process(event) ) {
        allFine = false;
      }   
    }

    if (!allFine) {
      cout << "Processing event # '" << iEvent << "' failed!" << endl;    
      cout << "HepMC record:" << endl;
      event->print();
    }

      // clean up before the next event
    jetFinder.ClearEvent(event);
    delete event;
  } 
  
    // a little bit of Pythia6 printout
  call_pystat(1);    
 
    // save the histograms 
  TFile *outFile = new TFile(config.OutputFileName.c_str(), "RECREATE");
  for (vector<baseAnalysis*>::iterator a = analyses.begin(); a != analyses.end(); ++a) {
    assert(*a);
    (*a)->Finalize(outFile);
  }
  
    // clean up 
  outFile->Close();
  delete outFile;

  for (vector<baseAnalysis*>::iterator a = analyses.begin(); a != analyses.end(); ++a) {
    delete (*a);
  }
  analyses.clear();

  cout << "pythia6: program ends" << endl;
  cout << "Successful run!" << endl;

  return 0;
}

