/**
 * Pythia8 example program.
 * 
 * Typical usage:
 *   ./pythia8.exe Process.config
 *
 * Created: by sj and ca
 * Last update (08.2012), Sebastian Piec
 */

#include <iostream>
#include <vector>

#include "Pythia.h"
#include "LHAFortran.h"
#include "HepMCInterface.h"
#include "HepMC/GenEvent.h"

using namespace Pythia8; 

#include "HepMC/IO_GenEvent.h"
#include "HepMC/GenEvent.h"
#include "HepMC/IO_AsciiParticles.h"
#include "HepMC/SimpleVector.h"

#include <TFile.h>

// HepMCAnalysis headers
#include "../../include/Configuration.h"
#include "../../include/JetFinder.h"

using namespace std;

// ---------------------------------------------------------------------- 
int main(int argc, const char* argv[]) 
{
  if (argc < 2) {
    cout << "Usage: " << argv[0] << " <Configuration File>" << endl;
    return 1;
  }
 
    // load job's configuration 
  Configuration config;
  if (!config.read(argv[1])) {
    cout << " ... skip !" << endl;
    return 1;
  }
  
  JetFinder jetFinder;
  jetFinder.Init(config.max_eta,config.min_pt);
  jetFinder.InitJetFinder(config.coneRadius, config.overlapThreshold, config.jet_ptmin,
      config.lepton_ptmin, config.DeltaR_lepton_track, config.jetalgorithm );

    // instantiate analyses requested by a user
  vector<baseAnalysis*> analyses = config.instantiateAnalyses();

  for (std::vector<baseAnalysis*>::const_iterator a = analyses.begin(); a != analyses.end(); ++a) {
    (*a)->Init(config.max_eta,config.min_pt);
    (*a)->InitJetFinder(config.coneRadius, config.overlapThreshold, config.jet_ptmin,
        config.lepton_ptmin, config.DeltaR_lepton_track, config.jetalgorithm);
  }

  HepMC::I_Pythia8 ToHepMC;

  // Generator. Process selection. LHC initialization.
  Pythia pythia(getenv("Pythia8DATA"));

  char processline[64];

  // Read in commands from external file.
  for (unsigned int files=0; files<config.ConfigFileNames.size(); files++)
      {
	  std::cout << "pythia8: Will use settings from file " << config.ConfigFileNames[files].c_str() << std::endl;
	  pythia.readFile(config.ConfigFileNames[files].c_str());
      }

  // as done for the random seed, if it is specified
  pythia.readString("Random:setSeed = on");
  sprintf(processline,"Random:seed = %ld",config.rseed);
  pythia.readString(processline);

  // master switches for FSR, ISR and MI
  sprintf(processline,"PartonLevel:FSR = %d",int(config.FSR));
  pythia.readString(processline);

  sprintf(processline,"PartonLevel:ISR = %d",int(config.ISR));
  pythia.readString(processline);

  sprintf(processline,"PartonLevel:MI = %d",int(config.MI));
  pythia.readString(processline);

  // Settings can also be changed with this command
  //pythia.readString("WeakBosonAndParton:all = on");

  pythia.init(2212, 2212, 14000.0);

  // Shorthand for the event and for settings.
  Event& event = pythia.event;
  Settings& settings = pythia.settings;

    // Extract settings to be used in the main program.
  int nEvent = config.nevents;

  //sj//  int nEvent = settings.mode("Main:numberOfEvents");
  //int nList = settings.mode("Main:numberToList");
  int nShow = settings.mode("Main:timesToShow");
  int nAbort = settings.mode("Main:timesAllowErrors");
  bool showChangedSettings = settings.flag("Main:showChangedSettings");
  bool showAllSettings = settings.flag("Main:showAllSettings");
  bool showAllParticleData = settings.flag("Main:showAllParticleData");

  // List changed data.
  if (showChangedSettings) settings.listChanged();
  if (showAllSettings) settings.listAll();

  // List particle data.  
  //if (showAllParticleData) ParticleDataTable::listAll();
  //starting at least for Pythia8 135 use the following line
  if (showAllParticleData) pythia.particleData.listAll();

  // Begin event loop.
  int nShowPace = max(1,nEvent/nShow); 
  int iAbort = 0; 
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

        if (iEvent%nShowPace == 0) 
	    cout << "pythia8: Now begin event "  << iEvent << endl;

    // Generate events. Quit if too many failures.
    if (!pythia.next()) {
      if (++iAbort < nAbort) continue;
      cout << "pythia8:  Event generation aborted prematurely, owing to error!\n"; 
      break;
    }
 
    // List first few events, both hard process and complete events.
    //if (iEvent < nList) { 
    //if (iEvent < nEvent) { 
    //  pythia.process.list();
      //event.list();
    // }

    HepMC::GenEvent* hepmcevt = new HepMC::GenEvent();
    ToHepMC.fill_next_event( event, hepmcevt );

    //hepmcevt->print();

      // call the jet finder
    vector<fastjet::PseudoJet> inclusiveJets = jetFinder.GetJet(hepmcevt);

      // run analyses
    bool allFine = true;
    for (vector<baseAnalysis*>::iterator a = analyses.begin(); a != analyses.end(); ++a) {
      assert(*a);
      (*a)->SetJet(&inclusiveJets);

      if ( !(*a)->Process(hepmcevt)) { 
        allFine = false;
      }
    }
    
    if (!allFine) {
      cout << "Processing event # '" << iEvent << "' failed!" << endl;
      cout << "HepMC record:" << endl;
      hepmcevt->print();
    }

      // clean up before the next event
    jetFinder.ClearEvent(hepmcevt);
    delete hepmcevt;
  }

    // print Pythia8 statistics
  pythia.statistics();

  pythia.settings.writeFile("allSettings.cmnd", true);

    // save the histograms
  TFile *outFile = new TFile(config.OutputFileName.c_str(), "RECREATE");
  for (vector<baseAnalysis*>::iterator a = analyses.begin(); a != analyses.end(); ++a) {
    (*a)->Finalize(outFile);
  }

    // clean up 
  outFile->Close();
  delete outFile;

  for (vector<baseAnalysis*>::iterator a = analyses.begin(); a != analyses.end(); ++a) {
    delete (*a);
  }
  analyses.clear();

  cout << "pythia8: program ends" << endl;
  cout << "Successful run!" << endl;
  
  return 0;
}

