#ifndef BASE_ANALYSIS_H_ 
#define BASE_ANALYSIS_H_ 

#include <iostream>
#include <string>
#include <map>
#include <vector>

#include "HepMC/GenEvent.h"

#include "fastjet/PseudoJet.hh"
#include "fastjet/JetDefinition.hh"
#include "fastjet/ClusterSequence.hh"

// forward declarations
//namespace fastjet {
//  class ClusterSequence;
//}
namespace HepPDT {
  class ParticleDataTable;
}
class TFile;
class TH1D;

//sj//enum RETURN{FAILURE=-1, SUCCESS=1, CHARGED=2, NOTCHARGED=3, FINALSTATEPARTICLE=4};

/**
@class baseAnalysis.h
@brief Base class for all  HepMCanalysis classes to implement common 
       functionality and interfaces.

@author Cano Ay Dec 2008 */
class baseAnalysis
{
public:
  baseAnalysis();
  virtual ~baseAnalysis();
 
  virtual int Finalize();
  virtual int Finalize(TFile* output);
  //virtual TH1D* popHisto(std::vector<TH1D*>);
  virtual TH1D* popHisto();

  virtual int Init(double maxeta = 2.5, double minpt = 0.5 );
  virtual int Process(HepMC::GenEvent* hepmcevt);
  //virtual std::vector<TH1D*> averagedHistograms() {TH1D* temp; std::vector<TH1D*> temp_vec; 
  //temp_vec.clear(); temp_vec.push_back(temp); return temp_vec;} ;

  virtual std::vector<TH1D*> averagedHistograms();
  
  virtual std::vector<TH1D*>& Histograms();

  /** Initialize FastJet Algorithm*/
  int InitJetFinder(double coneRadius, double overlapThreshold, double jet_ptmin, double lepton_ptmin, double DeltaR_lepton_track,
      int jetalgorithm = 2);

  int FindJetableParticles(HepMC::GenEvent* hepmcevt); 
  int FindJet(HepMC::GenEvent* hepmcevt);
  int DeleteJetObject(HepMC::GenEvent* hepmcevt);
  
  std::vector<fastjet::PseudoJet> GetJet(HepMC::GenEvent* hepmcevt);
  virtual void SetJet(std::vector<fastjet::PseudoJet>* inclusive_jets);

  /** Calculate Missing Et*/
  int FindMissingEt(HepMC::GenEvent* hepmcevt);
  int ClearMissingEt(HepMC::GenEvent* hepmcevt);

  /** Clear some values from the event */
  int ClearEvent(HepMC::GenEvent* hepmcevt);

  /** Set the Output filename*/  
  inline int setOutputFileName(const std::string& filename); 
  
  /** Set the directory name in Output root file*/  
  inline int setOutputRootDir(const std::string& dirname);

  /** Check some special neutral Particles */
  static bool IsNeutrino(int pid);
  static bool IsGamma(int pid);
  static bool IsNeutron(int pid);
  static bool IsK0(int pid); 
  static bool IsPi0(int pid);
  static bool IsElectron(int pid);
  static bool IsMuon(int pid);
  static bool IsGluon(int pid);

  /** Check if neutral particle*/  
  static bool chargedParticle(int pid);
  
  /** Check if final state particle*/
  virtual int IsFinalStateParticle(HepMC::GenParticle *p);

  /** Set the maximum allowed eta range*/  
  inline int setMaxEtaCut(double maxeta);
  
  /** Set maximum pt of tracks */  
  inline int setMinPtCut(double minpt);
  
  /** Initialization of histograms  */ 
  TH1D* initHist(const std::string &name, const std::string &title, const std::string &xlabel, 
      int nrBins=100, double xmin=0.0, double xmax=100.0);
  TH1D* initHistVariableBin(const std::string &name, const std::string &title, const std::string &xlabel, 
      int nbin, double nbinRange[]);
  
  /** check if mother decayed into daugther */ 
  bool checkDaughter(HepMC::GenParticle* mother, HepMC::GenParticle* daughter, int maxGenerations=-1);
  
  /** check if the track comes from a specific particle e.g. pid(W-Boson)=24 */ 
  bool trackfromPID(int pid, HepMC::GenParticle* track, int maxGenerations=-1);

  /** calculate the rapidity of a particle */
  static double getRapidity(const HepMC::GenEvent::particle_const_iterator& p);

protected:
  /** jet finding */
  fastjet::JetDefinition::Plugin*   m_plugin;
  fastjet::JetDefinition*           m_jetDef;

  std::vector< fastjet::PseudoJet >  m_input_particles;
  std::vector< fastjet::PseudoJet > m_inclusive_jets;
  fastjet::ClusterSequence* m_clust_seq;

  //std::map< std::string, int > m_histCounter;
  std::vector<TH1D*> m_histVector;
  //std::vector<TH1D*> m_histVectorVariableBin;  
  std::string m_outputFileName;
  std::string m_outputRootDir;

  //for jet finder
  double m_coneRadius;
  double m_overlapThreshold;
  double m_jet_ptmin;
  double m_lepton_ptmin;
  double m_DeltaR_lepton_track;
  bool m_Jetfinder_enabled;
  
  // specify the maximum eta of tracks
  double m_max_eta;
  double m_min_pt;

  // for calculation of the missing energy
  double exMissTruth;
  double eyMissTruth;
  double etMissTruth;
  double etsumMissTruth;

  size_t m_NanNum;
};

#endif      // BASE_ANALYSIS_H_ 

