/** 
 *
 * Last update (08.2012): Sebastian Piec
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <cstdlib>

#include "../include/Configuration.h"

#include "../include/baseAnalysis.h"
#include "../include/ttbarAnalysis.h"
#include "../include/ZtautauAnalysis.h"
#include "../include/WtaunuAnalysis.h"
#include "../include/bbbarAnalysis.h"
#include "../include/JetAnalysis.h"
#include "../include/WplusJetAnalysis.h"
#include "../include/ZAnalysis.h"
#include "../include/UEAnalysis.h"
#include "../include/EtMissAnalysis.h"
#include "../include/ElasScatAnalysis.h"
#include "../include/UserAnalysis.h"


using namespace std;

// ---------------------------------------------------------------------- 
Configuration::Configuration()
{
  useDefaultSettings();
}

// ---------------------------------------------------------------------- 
Configuration::Configuration(const string &configFile)
{
  useDefaultSettings();
  
  IsOK = read(configFile);
}

// ---------------------------------------------------------------------- 
Configuration::~Configuration()
{}

// ---------------------------------------------------------------------- 
void Configuration::useDefaultSettings()
{
    // the default randomseed will be generated from the time
  srand(time(0) % 900000000);
  rseed = rand();
  
    // default: 1000 events
  nevents = 1000;
  
    // default Outputfilename
  OutputFileName = "output.root";
  
    // switch off all analyses by default
  jet       = false;
  z         = false;
  wplusjet  = false;
  ztautau   = false;
  ttbar     = false;
  ue        = false;
  etmiss    = false;
  elasScat  = false;
  wtaunu    = false;
  bbbar     = false;
  user      = false;
  
    // default Jet parameters
  coneRadius          = 0.4;
  overlapThreshold    = 0.75;
  jet_ptmin           = 15.0;
  lepton_ptmin        = 15.0;
  DeltaR_lepton_track = 0.4;
  max_eta             = 2.5;
  min_pt              = 0.5;
  jetalgorithm        = 2;        // siscone = 1, antikt = 2
  
  FSR = true;
  ISR = true;
  MI  = true;
  
  IsOK = true;
}

/**
* @return logical true if configuration file reading was successful.
*/ 
bool Configuration::read( const string &configFile )
{
  ifstream myfile(configFile.c_str());
  cout << "Configuration::read(): loading configuration from file '" << configFile << "'" << endl;
  
  if (!myfile.is_open()) {
    cerr << "Configuration::read(): unable to open the file!" << endl;
    return false;
  }

    // clean-up before	    
  ConfigFileNames.clear();
  InputFileNames.clear();

    // read line by line of the file
  string line;
  while ( !myfile.eof() ) {
    string parameter, str1;
    
    int     mylong  = -9999;
    double  myfloat = -9999.9;
    bool    mybool  = false;
    
    // write read lines into string line
    getline( myfile, line );
    
    // now split the line into the subwords
    istringstream instring( line );
    istringstream instring2( line );
    istringstream instring3( line );
    
    // and fill string or float
    instring >> parameter >> str1 ;
    instring2 >> parameter >> myfloat ;
    instring3 >> parameter >> mybool ;
    mylong = static_cast< long >( myfloat );
    
    // check for Comments. If the line starts with # than skip the line
    string::size_type pos = parameter.find( "#" );
    if ( pos != string::npos ) continue;
    
    // check for Comments. If the line starts with * than skip the line
    string::size_type pos2 = parameter.find( "*" );
    if ( pos2 != string::npos ) continue;
    
    // check for Comments. If the line starts with '/' than skip the line
    string::size_type pos3 = parameter.find( "/" );
    if( pos3 != string::npos ) continue;
    
    // skip if the string is empty
    if( parameter.empty() ) continue;
    
    // read pysub entries
    if ( ( parameter == "rseed" ) && ( mylong != -9999 ) ) {
      rseed = mylong;
      cout << "readconfigFile: Parameter " << parameter << "=" << rseed << endl;
    }
    else if ( ( parameter == "nevents" ) && ( mylong != -9999 ) ) {
      nevents = static_cast< unsigned int >( mylong );
      cout << "readconfigFile: Parameter " << parameter << "=" << nevents << endl;
    }
    else if ( ( parameter == "OutputFileName" ) && ( !str1.empty() ) ) {
      OutputFileName = str1;
      cout << "readconfigFile: Parameter " << parameter << "=" << OutputFileName << endl;
    }
    else if ( ( parameter == "ConfigFileNames" ) && ( !str1.empty() ) ) {
      ConfigFileNames.push_back( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << ConfigFileNames.back() << endl;
    }
    else if ( ( parameter == "InputFileNames" ) && ( !str1.empty() ) ) {
      InputFileNames.push_back( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << InputFileNames.back() << endl;
    }
    else if ( ( parameter == "jet_analysis" ) && ( !str1.empty() ) ) {
      jet = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << jet << endl;
    }
    else if ( ( parameter == "z_analysis" ) && ( !str1.empty() ) ) {
      z = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << z << endl;
    }
    else if ( ( parameter == "wplusjet_analysis" ) && ( !str1.empty() ) ) {
      wplusjet = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << wplusjet << endl;
    }
    else if ( ( parameter == "ztautau_analysis" ) && ( !str1.empty() ) ) {
      ztautau = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << ztautau << endl;
    }
    else if ( ( parameter == "ttbar_analysis" ) && ( !str1.empty() ) ) {
      ttbar = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << ttbar << endl;
    }
    else if ( ( parameter == "ue_analysis" ) && ( !str1.empty() ) ) {
      ue = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << ue << endl;
    }
    else if ( ( parameter == "etmiss_analysis" ) && ( !str1.empty() ) ) {
      etmiss = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << etmiss << endl;
    }
    else if ( ( parameter == "elasScat_analysis" ) && ( !str1.empty() ) ) {
      elasScat = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << elasScat << endl;
    }
    else if ( ( parameter == "wtaunu_analysis" ) && ( !str1.empty() ) ) {
      wtaunu = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << wtaunu << endl;
    }
    else if ( ( parameter == "bbbar_analysis" ) && ( !str1.empty() ) ) {
      bbbar = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << bbbar << endl;
    }
    else if ( ( parameter == "user_analysis") && ( !str1.empty() ) ) {
      user = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << user << endl;
    }
    else if ( ( parameter == "jet_coneRadius" ) && ( myfloat != -9999.9 ) ) {
      coneRadius = myfloat;
      cout << "readconfigFile: Parameter " << parameter << "=" << coneRadius << endl;
    }
    else if ( ( parameter == "jet_overlapThreshold" ) && ( myfloat != -9999.9 ) ) {
      overlapThreshold = myfloat;
      cout << "readconfigFile: Parameter " << parameter << "=" << overlapThreshold << endl;
    }
    else if ( ( parameter == "jet_ptmin" ) && ( myfloat != -9999.9 ) ) {
      jet_ptmin = myfloat;
      cout << "readconfigFile: Parameter " << parameter << "=" << jet_ptmin << endl;
    }
    else if ( ( parameter == "DeltaR_lepton_track" ) && ( myfloat != -9999.9 ) ) {
      DeltaR_lepton_track = myfloat;
      cout << "readconfigFile: Parameter " << parameter << "=" << DeltaR_lepton_track << endl;
    }
    else if ( ( parameter == "lepton_ptmin" ) && ( myfloat != -9999.9 ) ) {
      lepton_ptmin = myfloat;
      cout << "readconfigFile: Parameter " << parameter << "=" << lepton_ptmin << endl;
    }
    else if ( ( parameter == "max_eta" ) && ( myfloat != -9999.9 ) ) {
      max_eta = myfloat;
      cout << "readconfigFile: Parameter " << parameter << "=" << max_eta << endl;
    }
    else if ( ( parameter == "min_pt" ) && ( myfloat != -9999.9 ) ){
      min_pt = myfloat;
      cout << "readconfigFile: Parameter " << parameter << "=" << min_pt << endl;
    }
    else if ((parameter == "jetalgorithm") && (mylong!=-9999)){
      jetalgorithm=mylong;
      cout << "readconfigFile: Parameter " << parameter << "=" << jetalgorithm << endl;
    }
    else if ( ( parameter == "FSR" ) && ( !str1.empty() ) ) {
      FSR = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << FSR << endl;
    }
    else if ( ( parameter == "ISR" ) && ( !str1.empty() ) ) {
      ISR = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << ISR << endl;
    }
    else if ( ( parameter == "MI" ) && ( !str1.empty() ) ) {
      MI = isParamOn( str1 );
      cout << "readconfigFile: Parameter " << parameter << "=" << MI << endl;
    }
    else{
      cout << "readconfigFile: Parameter " << parameter << " not recognized!" << endl;
      return false;
    }
  }
  myfile.close();
  
  return true;
}

/**
 * Create and initialize vector of analysis objects (kind of factory DP...)
 */
vector<baseAnalysis*> Configuration::instantiateAnalyses() const
{
  vector<baseAnalysis*> analyses;

  if (this->jet) {
    cout << "Adding module JetAnalysis" << endl;
    analyses.push_back(new JetAnalysis);
  }

  if (this->wplusjet) {
    cout << "Adding module WplusJetAnalysis" << endl;
    analyses.push_back(new WplusJetAnalysis);
  }

  if (this->z) {
    cout << "Adding module ZAnalysis" << endl;
    analyses.push_back(new ZAnalysis);
  }

  if (this->ztautau) {
    cout << "Adding module ZtautauAnalysis" << endl;
    analyses.push_back(new ZtautauAnalysis);
  }

  if (this->wtaunu) {
    cout << "Adding module WtaunuAnalysis" << endl;
    analyses.push_back(new WtaunuAnalysis);
  }

  if (this->bbbar) {
    cout << "Adding module bbbarAnalysis" << endl;
    analyses.push_back(new bbbarAnalysis);
  }

  if (this->ttbar) {
    cout << "Adding module ttbarAnalysis" << endl;
    analyses.push_back(new ttbarAnalysis);
  }

  if (this->ue) {
    cout << "Adding module UEAnalysis" << endl;
    analyses.push_back(new UEAnalysis);
  }  

  if (this->etmiss) {
    cout << "Adding module EtMissAnalysis" << endl;
    analyses.push_back(new EtMissAnalysis);
  }

  if (this->elasScat) {
    cout << "Adding module ElasScatAnalysis" << endl;
    analyses.push_back(new ElasScatAnalysis);
  }

  if (this->user) {
    cout << "Adding module UserAnalysis" << endl;
    analyses.push_back(new UserAnalysis);
  }

  return analyses;
}

/** 
 * @return logical true if a given parameter is switched ON. 
 */ 
bool Configuration::isParamOn(const string &param) const
{
    // convert to lower case
  string paramLow = param;
  transform(paramLow.begin(), paramLow.end(), paramLow.begin(), ::tolower);

  return (paramLow == "on") || (paramLow == "true");
}

