/**
 * Analysis example for elastic scattering with PYTHIA8 (provided by Hasko Stenzel)
 *
 * (Refactored on 9.2012, SP)
*/

#include <iostream>
#include <stdio.h>

#include "HepMC/GenEvent.h"
#include "HepMC/IO_GenEvent.h"
#include "HepMC/GenParticle.h"
#include "HepMC/GenVertex.h"
#include "HepMC/IO_AsciiParticles.h"
#include "HepMC/SimpleVector.h"
#include "CLHEP/Vector/LorentzVector.h"

#include "TH1.h"
#include "TMath.h"
#include "TLorentzVector.h"
#include "TFile.h"

#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"
#include "fastjet/JetDefinition.hh"
#include "fastjet/SISConePlugin.hh"

#include "../include/ElasScatAnalysis.h"

using namespace std;

// ---------------------------------------------------------------------- 
ElasScatAnalysis::ElasScatAnalysis()
{}

// ---------------------------------------------------------------------- 
ElasScatAnalysis::~ElasScatAnalysis()
{}

// ---------------------------------------------------------------------- 
int ElasScatAnalysis::Init(double maxEta, double minPt)
{
  m_max_eta = maxEta;
  m_min_pt = minPt;
  
  m_outputFileName = "ElasScatAnalysis.root";
  m_outputRootDir = "ElasScat";

    // book histograms 
  n_proton = initHist("nproton", "nproton", "Number of Protons", 5, -0.5, 4.5);
  p1_pt = initHist("Ptp1", "p_{T} proton1", " p_{T} [GeV]", 100, 0.0, 1.0);
  p1_eta = initHist("eta1", "#eta 1st proton", "#eta", 200, 9.0, 16.0);
  p1_theta = initHist("theta1", "sacttering angle p1", "#theta [#mu rad]", 200, 0.0, 100.0);
  p1_phi = initHist("p1_phi", "#varphi p1", "#varphi", 64, -3.2, 3.2);

  p2_pt = initHist("Ptp2", "p_{T} proton2", " p_{T} [GeV]", 100, 0.0, 1.0);
  p2_eta = initHist("eta2", "#eta 2nd proton", "#eta", 200, -16.0, -9.0);
  p2_theta = initHist("theta2", "sacttering angle p2", "#theta [#mu rad]", 200, -100.0, 0.0);
  p2_phi = initHist("p2_phi", "#varphi p2", "#varphi", 64, -3.2, 3.2);

    // two histograms for the t-spectrum
  t_spect = initHist("tspect", "t-spectrum", "t", 100, 0.0, 0.1);
  tau_spect = initHist("tauspect", "#tau-spectrum", "#tau=log_{10}|t|", 100, -5.0, 0.0);

  // beam proton histograms (to check beam class settings)
  pi1_en = initHist("pi1en", "Energy deviation", " E-E_{0} [GeV]", 100, -2.0, 2.0);
  pi1_divx = initHist("pi1divx", "Divergence in x beam 1", "angle [#mu rad]", 100, -2.0, 2.0);
  pi1_divy = initHist("pi1divy", "Divergence in y beam 1", "angle [#mu rad]", 100, -2.0, 2.0);

  pi2_en = initHist("pi2en", "Energy deviation", " E-E_{0} [GeV]", 100, -2.0, 2.0);
  pi2_divx = initHist("pi2divx", "Divergence in x beam 2", "angle [#mu rad]", 100, -2.0, 2.0);
  pi2_divy = initHist("pi2divy", "Divergence in y beam 2", "angle [#mu rad]", 100, -2.0, 2.0);

  return true; 
}

// ---------------------------------------------------------------------- 
int ElasScatAnalysis::Process(HepMC::GenEvent *event)
{ 
  int pcount=0;
  int pint=0;
  const double plhc = 7000.0;
  
  HepMC::GenParticle *pi1 = 0;
  HepMC::GenParticle *pi2 = 0;
  HepMC::GenParticle *p1 = 0;
  HepMC::GenParticle *p2 = 0;

    // loop over the final state particles 
  for (HepMC::GenEvent::particle_const_iterator p = event->particles_begin(); p != event->particles_end(); ++p) {
    int pid = (*p)->pdg_id();
      
    if ( (event->beam_particles().first->pdg_id() == 2212 && !(*p)->production_vertex()) || 
         (event->beam_particles().second->pdg_id() == 2212 && !(*p)->production_vertex()) ) {

	    if (pid == 2212) { 
    	  ++pint;

	      if (pint == 1) {
          pi1=(*p);
        }

    	  if (pint == 2) {
          pi2=(*p);
        }
	    } 
    }

    if ( ((*p)->status()%1000 == 1) && (!(*p)->end_vertex()) )  {
	    if (pid == 2212) { 
    	  pcount++;
	      
        if (pcount == 1) {
          p1=(*p);
        }
    	  
        if (pcount == 2) {
          p2=(*p);
        }

	      if (pcount > 2) {
          cout << "Strange: More than two protons in this event!" << "\n";
        }
  	  }
    }
  }

  if (pint == 2) {
    pi1_en->Fill((pi1->momentum().e())-plhc);
    double divx = (pi1->momentum().px())/(pi1->momentum().pz());
    double divy = (pi1->momentum().py())/(pi1->momentum().pz());
    pi1_divx->Fill(divx*1.E6);
    pi1_divy->Fill(divy*1.E6);

    pi2_en->Fill((pi2->momentum().e())-plhc);
    divx = (pi2->momentum().px())/(pi2->momentum().pz());
    divy = (pi2->momentum().py())/(pi2->momentum().pz());
    pi2_divx->Fill(divx*1.E6);
    pi2_divy->Fill(divy*1.E6);
  }
    
  if (pcount == 2) {
    n_proton->Fill(pcount);
    p1_pt->Fill(p1->momentum().perp());
    p1_eta->Fill(p1->momentum().eta());
    p1_phi->Fill(p1->momentum().phi());
    p1_theta->Fill((p1->momentum().theta())*1.E6); //micro-radians

    p2_pt->Fill(p2->momentum().perp());
    p2_eta->Fill(p2->momentum().eta());
    p2_phi->Fill(p2->momentum().phi());
    p2_theta->Fill((p2->momentum().theta()-M_PI)*1.E6); //micro-radians

     // t-calculation 
    double st1 =(p1->momentum().theta())*plhc;
    double st2 =(p2->momentum().theta()-M_PI)*plhc;
    double t1 = st1*st1; 
    double t2 = st2*st2; 
    double tmean = (t1+t2)/2.; // average t-value reconstructed from the scattering angles

    double tau = log10(tmean);
    t_spect->Fill(tmean); 
    tau_spect->Fill(tau); 
  }
  
  return true;
}
 
