/**
 *
 * (Refactored on 09.2012, SP) 
*/

#include <iostream>

#include "HepMC/GenEvent.h"
#include "HepMC/IO_GenEvent.h"
#include "HepMC/GenParticle.h"
#include "HepMC/GenVertex.h"
#include "HepMC/IO_AsciiParticles.h"
#include "CLHEP/Vector/LorentzVector.h"

#include "TH1.h"
#include "TLorentzVector.h"

#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"
#include "fastjet/JetDefinition.hh"
#include "fastjet/SISConePlugin.hh"

#include "../include/JetAnalysis.h"

using namespace std;

// ---------------------------------------------------------------------- 
JetAnalysis::JetAnalysis()
{}

// ---------------------------------------------------------------------- 
JetAnalysis::~JetAnalysis()
{}

// ---------------------------------------------------------------------- 
int JetAnalysis::Init(double maxEta, double minPt)
{
  m_max_eta = maxEta;       // default eta cut
  m_min_pt = minPt;         // default pt cut
   
  m_outputFileName = "JetAnalysis.root";
  m_outputRootDir = "Jet";
  
    // book histograms
  m_jet_count = initHist("jet_count", "Nr of Jets", "Nr Jets", 16, -0.5, 15.5);
  m_jet_pt = initHist("jet_pt", "Jet Transverse Momentum", "Jet p_{T} [GeV]", 40, 0.0, 400.0);
  m_jet_pt_log = initHist("jet_pt_logscale", "Jet Transverse Momentum -logscale-", "Jet p_{T} [GeV]", 50, 0.0, 150.0);

  m_leadingjet_pt = initHist("leadingjet_pt", "Leading Jet Transverse Momentum", "Jet p_{T} [GeV]", 50, 0.0, 150.0);
  m_leadingjet_pt_log = initHist("leadingjet_pt_logscale", "Leading Jet Transverse Momentum -logscale-",
      "Jet p_{T} [GeV]", 50, 0.0, 150.0);

  m_secondleadingjet_pt = initHist("secondleadingjet_pt", "Second leading Jet Transverse Momentum", 
      "Jet p_{T} [GeV]", 50, 0.0, 150.0);
  m_secondleadingjet_pt_log = initHist("secondleadingjet_pt_logscale", "Second leading Jet Transverse Momentum -logscale-", 
      "Jet p_{T} [GeV]", 50, 0.0, 150.0);
  
  m_jet_phi = initHist("jet_phi", "Jet Phi angle", "Jet #varphi", 64, 0.0, 6.4);
  m_jet_eta = initHist("jet_eta", "Jet Pseudo rapidity", "Jet #eta", 60, -6.0, 6.0);
  
  m_charged_particle_multiplicity = initHist("charged_particle_multiplicity", "Number of charged Particles in the Event with #eta <2.5 and p_{T}>0.5 GeV",
      "charged particle multiplicity", 150, 0.0, 150.0);
  
  m_charged_particle_pt = initHist("charged_particle_pt","p_{T} of charged particles with #eta <2.5 and p_{T}>0.5 GeV","p_{T}", 
      15, 0.0, 15.0); 
  m_charged_particle_pt_log = initHist("charged_particle_pt_logscale","p_{T} of charged particles with #eta <2.5 and p_{T}>0.5 GeV","p_{T}", 
      15, 0.0, 15.0);
   
  m_charged_particle_mean_pt = initHist("charged_particle_mean_pt", "Average p_{T} charged Particles in the Event with #eta <2.5 and p_{T}>0.5 GeV",
      "charged particle #bar{p}_{T}", 40, 0.0, 10.0);
  m_charged_particle_rms_pt = initHist("charged_particle_rms_pt", "RMS p_{T} charged Particles in the Event with #eta <2.5 and p_{T}>0.5 GeV",
      "charged particle #sigma(p_{T})", 40, 0.0, 10.0);
  
  m_jet_Deta_leadingJet_secondJet = initHist("jet_Deta_leadingJet_secondJet", "#Delta#eta leading and second leading Jet", "#Delta#eta", 
      60, 0.0, 6.0);
  m_jet_Dphi_leadingJet_secondJet = initHist("jet_Dphi_leadingJet_secondJet", "#Delta#varphi leading and second leading Jet", "#Delta#varphi", 
      64, 0.0, 3.2);
  m_jet_DR_leadingJet_secondJet = initHist("jet_DR_leadingJet_secondJet", "#DeltaR leading and second leading Jet", "#DeltaR", 
      50, 0.0, 10.0);
  
  return true;
}

// ---------------------------------------------------------------------- 
/** Main analysis function. */ 
// ---------------------------------------------------------------------- 
int JetAnalysis::Process(HepMC::GenEvent *event) 
{
  size_t nChargedParticles = 0;  

    // auto_ptr<> is deprecated in C++11. Use unique_ptr<> instead, when new gcc (~4.7) is availble... (SP)
  auto_ptr<TH1D> temp_pt(new TH1D("jetanalysis_temp_pt", "temp_pt", 200, 0.0, 200.0));
  
    // loop over the particles and select pdgid and pt
  for (HepMC::GenEvent::particle_const_iterator p = event->particles_begin(); p != event->particles_end(); ++p) {
    int pid = (*p)->pdg_id();       // get particle's PDG ID
      
      // remove neutrinos, gammas and neutrons from the final state
    if (!chargedParticle(pid)) {
      continue;
    }
      
      // take all stable particles of this event
    if (!IsFinalStateParticle((*p))) {
      continue;
    }
  
    const double momPerp = (*p)->momentum().perp();
    
    temp_pt->Fill(momPerp);
    m_charged_particle_pt->Fill(momPerp);
    m_charged_particle_pt_log->Fill(momPerp);
    
    ++nChargedParticles;
  }
  
    // storage of input particles for jet
  CLHEP::HepLorentzVector lv_leadingJet(0, 0, 0, 0);
  
  m_jet_count->Fill(m_inclusive_jets.size());
  m_charged_particle_multiplicity->Fill(nChargedParticles);
  
  m_charged_particle_mean_pt->Fill(temp_pt->GetMean());
  m_charged_particle_rms_pt->Fill(temp_pt->GetRMS());
  
  if (!m_inclusive_jets.empty()) {
    for (size_t i = 0; i < m_inclusive_jets.size(); ++i) {
      m_jet_pt->Fill(m_inclusive_jets[i].perp());
      m_jet_pt_log->Fill(m_inclusive_jets[i].perp());
    }

    m_leadingjet_pt->Fill(m_inclusive_jets[0].perp());
    m_leadingjet_pt_log->Fill(m_inclusive_jets[0].perp());
    
    if (m_inclusive_jets.size() > 1) {
      m_secondleadingjet_pt->Fill(m_inclusive_jets[1].perp());
      m_secondleadingjet_pt_log->Fill(m_inclusive_jets[1].perp());
    } else {
      //cout << "[JetAnalysis] WARNING: No second leading jet!" << endl; 
    }

    m_jet_phi->Fill(m_inclusive_jets[0].phi());
    m_jet_eta->Fill(m_inclusive_jets[0].eta());
    
    CLHEP::HepLorentzVector lv_firstJet(m_inclusive_jets[0].px(), m_inclusive_jets[0].py(), m_inclusive_jets[0].pz(), m_inclusive_jets[0].e());

    if (m_inclusive_jets.size() > 1) {
      CLHEP::HepLorentzVector lv_secondJet(m_inclusive_jets[1].px(), m_inclusive_jets[1].py(), m_inclusive_jets[1].pz(), m_inclusive_jets[1].e());   
      m_jet_DR_leadingJet_secondJet->Fill(fabs(lv_firstJet.deltaR(lv_secondJet)));
      m_jet_Deta_leadingJet_secondJet->Fill(fabs(lv_firstJet.eta()-lv_secondJet.eta()));
      m_jet_Dphi_leadingJet_secondJet->Fill(fabs(lv_firstJet.vect().deltaPhi(lv_secondJet.vect())));
    }
  }

  return true;
}

